import json
import re
from collections import defaultdict

from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from flask import current_app
from werkzeug.routing import IntegerConverter, NumberConverter


def generate_swagger(app, api_version, doc_path, format="yaml", **options):
    generator = SwaggerGenerator(
        app, api_version, doc_path, format, **options
    )
    generator.generate_swagger()


class SwaggerGenerator:
    def __init__(self, app, api_version, doc_path, format="yaml", **options):
        """
        :param app: flask app instance
        :param api_version: api doc version
        :param doc_path: swagger save path
        :param format:swagger file format, json or yaml
        """
        self.app = app
        self.api_version = api_version
        self.doc_path = doc_path.rstrip("/")
        self.doc_format = format
        self.api_spec_opts = options
        self.api_specs = {}
        self.blueprint_to_apis = defaultdict(list)

    def get_api_spec(self, name):
        """
        get api spec for give blueprint or app
        :param name: blurprint name or app.import_name
        :return: api_spec
        """
        if name not in self.api_specs:
            self.api_specs[name] = APISpec(
                title=f"{name}-api",
                version=self.api_version,
                openapi_version="3.0.0",
                plugins=(MarshmallowPlugin(),),
                **self.api_spec_opts
            )
        return self.api_specs[name]

    def generate_swagger(self):
        with self.app.app_context():
            for rule in self.app.url_map.iter_rules():
                self.register_rule(rule)
        for doc_name, apis in self.blueprint_to_apis.items():
            doc_path = f"{self.doc_path}/{doc_name}.{self.doc_format}"
            api_spec = self.get_api_spec(doc_name)
            for api in apis:
                api_spec.path(**api)
            with open(doc_path, "w") as f:
                if self.doc_format == "json":
                    json.dump(
                        api_spec.to_dict(), f, indent=4, ensure_ascii=False,
                        sort_keys=True
                    )
                else:
                    f.write(
                        api_spec.to_yaml(
                            yaml_dump_kwargs={"allow_unicode": True}
                        )
                    )

    def register_rule(self, rule):
        rule_parser = RuleParser(self.app, rule)
        operations = rule_parser.operations
        if not operations:
            return
        self.blueprint_to_apis[rule_parser.doc_file_name].append(
            dict(
                path=rule_parser.path,
                operations=operations,
                parameters=rule_parser.parameters,
                description=rule_parser.description
            )
        )


class RuleParser:
    def __init__(self, app, rule):
        self.app = app
        self.rule = rule
        self.view_func = self.app.view_functions[self.rule.endpoint]

    def get_blueprint_name(self):
        blueprint_name = self.rule.endpoint.rsplit(".", 1)[0]
        if blueprint_name in self.app.blueprints:
            return blueprint_name
        return self.app.import_name

    @property
    def doc_file_name(self):
        if current_app.config.get("REST_SERIALIZER_BLUEPRINT_SEPARATE_DOCS",
                                  False):
            view_cls = getattr(self.view_func, "view_class", None)
            if view_cls and getattr(view_cls, "rest_serialize_doc_name", None):
                return getattr(view_cls, "rest_serialize_doc_name", None)
            return self.get_blueprint_name()
        return self.app.import_name

    @property
    def description(self):
        return self.view_func.__doc__ or self.rule.endpoint

    @property
    def path(self):
        return re.sub(r"<(?:\S+:)?(\S+)>", r"{\1}", self.rule.rule)

    @property
    def parameters(self):
        parameters = []
        for argument in self.rule.arguments:
            convertor = self.rule._converters.get(argument)
            arg_type = "string"
            if convertor:
                if isinstance(convertor, IntegerConverter):
                    arg_type = "integer"
                elif isinstance(convertor, NumberConverter):
                    arg_type = "number"
            parameters.append({
                "name": argument,
                "in": "path",
                "required": True,
                "schema": {"type": arg_type}
            })

        return parameters

    @property
    def operations(self):
        view_cls = getattr(self.view_func, "view_class", None)
        if view_cls:
            operations = {}
            for method in ("get", "post", "delete", "put", "patch"):
                handler_func = getattr(view_cls, method, None)
                if handler_func:
                    request_schema = getattr(handler_func, "request_schema",
                                             None)
                    response_schema = getattr(handler_func,
                                              "response_schema", None)
                    method_operation = self._get_operations([method],
                                                            request_schema,
                                                            response_schema)
                    if method_operation:
                        operations.update(method_operation)
            return operations
        else:
            request_schema = getattr(self.view_func, "request_schema", None)
            response_schema = getattr(self.view_func, "response_schema", None)
            return self._get_operations(self.rule.methods, request_schema,
                                        response_schema)

    @staticmethod
    def _get_operations(methods, request_schema, response_schema):
        if not any([request_schema, response_schema]):
            return
        operations = {}
        methods = [method.lower() for method in methods if
                   method.lower() not in ("options", "head")]
        for method in methods:
            content = {
                "application/json": {
                    "schema": response_schema or {}
                }
            } if response_schema else {}
            operations[method] = dict(
                responses={
                    "200": {
                        "description": "SUCCESS",
                        "content": content
                    }
                }
            )
            if request_schema:
                if method in ("get", "delete"):
                    operations[method]["parameters"] = [
                        {"schema": request_schema, "in": "query"}
                    ]
                else:
                    operations[method]["requestBody"] = {
                        "content": {
                            "application/json": {
                                "schema": request_schema
                            }
                        }
                    }
        return operations



import functools
import logging

from flask import Response, current_app, jsonify, request
from marshmallow import Schema, ValidationError, missing
from werkzeug.exceptions import BadRequest

logger = logging.getLogger(__name__)


class RequestFullData(dict):
    def get_val_from_request(self, location, attr_name):
        missing_default = missing
        if location in ("args", "query_string"):
            return request.args.get(attr_name, missing_default)
        elif location == "headers":
            return request.headers.get(attr_name, missing_default)
        elif location == "body":
            if request.is_json:
                return request.json.get(attr_name, missing_default)
            return request.form.get(attr_name, )
        return missing_default

    def get(self, key, default=None):
        location, attr_name = key.split(".")
        return self.get_val_from_request(location, attr_name)


class RequestSerializer:
    @staticmethod
    def get_default_location(method):
        if method.lower() in ("get", "delete"):
            return "args"
        else:
            return "body"

    @staticmethod
    def serialize_request(request_schema):
        """serialize request information into python dict"""
        if request_schema is None:
            return {}

        field_origin_data_key = {}
        for field_name, field_obj in request_schema.load_fields.items():
            # save origin data_key and restore it after serialize request
            field_origin_data_key[field_name] = field_obj.data_key
            # get location for field
            # if missing, get a default according to http method
            location = field_obj.metadata.get("location") or \
                       RequestSerializer.get_default_location(request.method)
            # set location as data_key of field
            field_obj.data_key = f"{location}.{field_name}"

        serialized_data = request_schema.load(RequestFullData())

        for field_name, field_obj in request_schema.load_fields.items():
            field_obj.data_key = field_origin_data_key.get(field_name)
            if field_name not in serialized_data:
                # serialized_data is passed into view_function as kwargs,
                # make sure every key exists
                serialized_data[field_name] = None

        return serialized_data


def get_schema_instance(schema):
    if schema is not None:
        if isinstance(schema, Schema):
            return schema
        elif issubclass(schema, Schema):
            return schema()
    return None


def serialize_with_schemas(request_schema=None,
                           response_schema=None):
    """

    :param request_schema: instance or subclass of marshmallow Schema
    :param response_schema: instance or subclass of marshmallow Schema
    :return: a decorator
    """

    def deco(func):
        setattr(func, "request_schema", request_schema)
        setattr(func, "response_schema", response_schema)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            request_schema_instance = get_schema_instance(request_schema)
            response_schema_instance = get_schema_instance(response_schema)
            try:
                serialized_data = RequestSerializer.serialize_request(
                    request_schema_instance)
                kwargs.update(serialized_data)
            except ValidationError as e:
                logger.debug("serialize request with schema %s failed: %s",
                             request_schema_instance, e)
                get_validation_error_handler()(e)
            origin_ret = func(*args, **kwargs)
            if isinstance(origin_ret, Response) or \
                    response_schema_instance is None:
                return origin_ret
            return jsonify(response_schema_instance.dump(origin_ret))

        return wrapper

    return deco


def abort_handler(exc):
    raise BadRequest(description=str(exc.messages))


def re_raise_handler(exc):
    raise exc


validation_error_handlers = {
    "abort": abort_handler,
    "re_raise": re_raise_handler
}


def get_validation_error_handler():
    handler = current_app.config.get(
        "REST_SERIALIZER_VALIDATION_ERROR_HANDLER", "abort")
    if callable(handler):
        return handler
    return validation_error_handlers.get(handler, abort_handler)



from setuptools import setup

VERSION = "0.0.1"
with open("README.md") as f:
    readme = f.read()

setup(
    name="flask-rest-serializer",
    version=VERSION,
    description="Make it easy to serialize flask api request and response",
    long_description=readme,
    long_description_content_type='text/markdown',
    url="https://github.com/panjianjun/flask_rest_serializer",
    author="bayi",
    author_email="bayipan@gmail.com",
    license="Apache Software License",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    keywords="flask rest serializer automatically swagger",
    packages=["flask_rest_serializer"],
    install_requires=["flask", "marshmallow", "apispec", "pyyaml"],
)
